@@ -16,8 +16,8 @@ GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GCC; see the file COPYING.  If not, write to
-the Free Software Foundation, 59 Temple Place - Suite 330,
-Boston, MA 02111-1307, USA.  
+the Free Software Foundation, 51 Franklin Street, Fifth Floor,
+Boston, MA 02110-1301, USA.  
 
 Java and all Java-based marks are trademarks or registered trademarks
 of Sun Microsystems, Inc. in the United States and other countries.
@@ -62,8 +62,8 @@ The Free Software Foundation is independent of Sun Microsystems, Inc.  */
    open() call.
 
    If you want to add another host, add your define to the list below
-   (i.e. defined(WIN32) || defined(YOUR_HOST)) and add an host-specific
-   .c file to Make-lang.in similar to win32-host.c  */
+   (i.e. defined(WIN32) || defined(YOUR_HOST)) and add a host-specific
+   .c file to Make-lang.in similar to win32-host.c.  */
 #if defined(WIN32)
 extern int
 jcf_open_exact_case (const char* filename, int oflag);
