@@ -1,5 +1,5 @@
 /* RandomAccessFile.java -- Class supporting random file I/O
-   Copyright (C) 1998, 1999, 2001, 2002, 2003 Free Software Foundation, Inc.
+   Copyright (C) 1998, 1999, 2001, 2002, 2003, 2004, 2005  Free Software Foundation, Inc.
 
 This file is part of GNU Classpath.
 
@@ -15,8 +15,8 @@ General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU Classpath; see the file COPYING.  If not, write to the
-Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-02111-1307 USA.
+Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+02110-1301 USA.
 
 Linking this library statically or dynamically with other modules is
 making a combined work based on this library.  Thus, the terms and
@@ -38,8 +38,9 @@ exception statement from your version. */
 
 package java.io;
 
+import gnu.java.nio.channels.FileChannelImpl;
+
 import java.nio.channels.FileChannel;
-import java.nio.channels.FileChannelImpl;
 
 /* Written using "Java Class Libraries", 2nd edition, ISBN 0-201-31002-3
  * "The Java Language Specification", ISBN 0-201-63451-1
@@ -54,19 +55,19 @@ import java.nio.channels.FileChannelImpl;
  * <code>DataInput</code> and <code>DataOutput</code> interfaces to allow
  * the reading and writing of Java primitives.
  *
- * @author Aaron M. Renn <arenn@urbanophile.com>
- * @author Tom Tromey <tromey@cygnus.com>
+ * @author Aaron M. Renn (arenn@urbanophile.com)
+ * @author Tom Tromey (tromey@cygnus.com)
  */
 public class RandomAccessFile implements DataOutput, DataInput
 {
 
   // The underlying file.
+  private FileChannelImpl ch;
   private FileDescriptor fd;
   // The corresponding input and output streams.
   private DataOutputStream out;
   private DataInputStream in;
   
-  private FileChannel ch; /* cached associated file-channel */
   
   /**
    * This method initializes a new instance of <code>RandomAccessFile</code>
@@ -85,12 +86,46 @@ public class RandomAccessFile implements DataOutput, DataInput
    * illegal value
    * @exception SecurityException If the requested access to the file 
    * is not allowed
-   * @exception IOException If any other error occurs
+   * @exception FileNotFoundException If the file is a directory, or 
+   * any other error occurs
    */
   public RandomAccessFile (File file, String mode)
     throws FileNotFoundException
   {
-    this (file.getPath(), mode);
+    int fdmode;
+    if (mode.equals("r"))
+      fdmode = FileChannelImpl.READ;
+    else if (mode.equals("rw"))
+      fdmode = FileChannelImpl.READ | FileChannelImpl.WRITE;
+    else if (mode.equals("rws"))
+      {
+	fdmode = (FileChannelImpl.READ | FileChannelImpl.WRITE
+		  | FileChannelImpl.SYNC);
+      }
+    else if (mode.equals("rwd"))
+      {
+	fdmode = (FileChannelImpl.READ | FileChannelImpl.WRITE
+		  | FileChannelImpl.DSYNC);
+      }
+    else
+      throw new IllegalArgumentException ("invalid mode: " + mode);
+
+    final String fileName = file.getPath();
+
+    // The obligatory SecurityManager stuff
+    SecurityManager s = System.getSecurityManager();
+    if (s != null)
+      {
+        s.checkRead(fileName);
+
+        if ((fdmode & FileChannelImpl.WRITE) != 0)
+          s.checkWrite(fileName);
+      }
+
+    ch = new FileChannelImpl (file, fdmode);
+    fd = new FileDescriptor(ch);
+    out = new DataOutputStream (new FileOutputStream (fd));
+    in = new DataInputStream (new FileInputStream (fd));
   }
 
   /**
@@ -112,42 +147,13 @@ public class RandomAccessFile implements DataOutput, DataInput
    * illegal value
    * @exception SecurityException If the requested access to the file 
    * is not allowed
-   * @exception FileNotFoundException If any other error occurs
+   * @exception FileNotFoundException If the file is a directory or 
+   * any other error occurs
    */
   public RandomAccessFile (String fileName, String mode)
     throws FileNotFoundException
   {
-    int fdmode;
-    if (mode.equals("r"))
-      fdmode = FileDescriptor.READ;
-    else if (mode.equals("rw"))
-      fdmode = FileDescriptor.READ | FileDescriptor.WRITE;
-    else if (mode.equals("rws"))
-      {
-	fdmode = (FileDescriptor.READ | FileDescriptor.WRITE
-		  | FileDescriptor.SYNC);
-      }
-    else if (mode.equals("rwd"))
-      {
-	fdmode = (FileDescriptor.READ | FileDescriptor.WRITE
-		  | FileDescriptor.DSYNC);
-      }
-    else
-      throw new IllegalArgumentException ("invalid mode: " + mode);
-
-    // The obligatory SecurityManager stuff
-    SecurityManager s = System.getSecurityManager();
-    if (s != null)
-      {
-        s.checkRead(fileName);
-
-        if ((fdmode & FileDescriptor.WRITE) != 0)
-          s.checkWrite(fileName);
-      }
-
-    fd = new FileDescriptor (fileName, fdmode);
-    out = new DataOutputStream (new FileOutputStream (fd));
-    in = new DataInputStream (new FileInputStream (fd));
+    this (new File(fileName), mode);
   }
 
   /**
@@ -158,8 +164,7 @@ public class RandomAccessFile implements DataOutput, DataInput
    */
   public void close () throws IOException
   {
-    if (fd.valid())
-      fd.close();
+    ch.close();
   }
 
   /**
@@ -172,10 +177,12 @@ public class RandomAccessFile implements DataOutput, DataInput
    */
   public final FileDescriptor getFD () throws IOException
   {
-    if (! fd.valid())
-      throw new IOException ();
-
-    return fd;
+    synchronized (this)
+      {
+	if (fd == null)
+	  fd = new FileDescriptor (ch);
+	return fd;
+      }
   }
 
   /**
@@ -188,25 +195,39 @@ public class RandomAccessFile implements DataOutput, DataInput
    */
   public long getFilePointer () throws IOException
   {
-    return fd.getFilePointer();
+    return ch.position();
   }
 
   /**
-   * This method sets the length of the file to the specified length.  If
-   * the currently length of the file is longer than the specified length,
-   * then the file is truncated to the specified length.  If the current
-   * length of the file is shorter than the specified length, the file
-   * is extended with bytes of an undefined value.
-   *  <p>
+   * This method sets the length of the file to the specified length.
+   * If the currently length of the file is longer than the specified
+   * length, then the file is truncated to the specified length (the
+   * file position is set to the end of file in this case).  If the
+   * current length of the file is shorter than the specified length,
+   * the file is extended with bytes of an undefined value (the file
+   * position is unchanged in this case).
+   * <p>
    * The file must be open for write access for this operation to succeed.
    *
-   * @param newlen The new length of the file
+   * @param newLen The new length of the file
    *
    * @exception IOException If an error occurs
    */
   public void setLength (long newLen) throws IOException
   {
-    fd.setLength (newLen);
+    // FIXME: Extending a file should probably be done by one method call.
+
+    // FileChannel.truncate() can only shrink a file.
+    // To expand it we need to seek forward and write at least one byte.
+    if (newLen < length())
+      ch.truncate (newLen);
+    else if (newLen > length())
+      {
+	long pos = getFilePointer();
+	seek(newLen - 1);
+	write(0);
+	seek(pos);
+      }
   }
 
   /**
@@ -218,7 +239,7 @@ public class RandomAccessFile implements DataOutput, DataInput
    */
   public long length () throws IOException
   {
-    return fd.getLength ();
+    return ch.size();
   }
 
   /**
@@ -239,7 +260,7 @@ public class RandomAccessFile implements DataOutput, DataInput
    * bytes are stored starting at the beginning of the array and up to 
    * <code>buf.length</code> bytes can be read.
    *
-   * @param buf The buffer to read bytes from the file into
+   * @param buffer The buffer to read bytes from the file into
    *
    * @return The actual number of bytes read or -1 if end of file
    *
@@ -254,7 +275,7 @@ public class RandomAccessFile implements DataOutput, DataInput
    * This methods reads up to <code>len</code> bytes from the file into the
    * specified array starting at position <code>offset</code> into the array.
    *
-   * @param buf The array to read the bytes into
+   * @param buffer The array to read the bytes into
    * @param offset The index into the array to start storing bytes
    * @param len The requested number of bytes to read
    *
@@ -316,12 +337,12 @@ public class RandomAccessFile implements DataOutput, DataInput
    * significant byte first (i.e., "big endian") regardless of the native
    * host byte ordering. 
    * <p>
-   * As an example, if <code>byte1</code> and code{byte2</code> represent 
+   * As an example, if <code>byte1</code> and <code>byte2</code> represent 
    * the first
    * and second byte read from the stream respectively, they will be
    * transformed to a <code>char</code> in the following manner:
    * <p>
-   * <code>(char)(((byte1 & 0xFF) << 8) | (byte2 & 0xFF)</code>
+   * <code>(char)(((byte1 &amp; 0xFF) &lt;&lt; 8) | (byte2 &amp; 0xFF)</code>
    * <p>
    * This method can read a <code>char</code> written by an object 
    * implementing the
@@ -397,7 +418,7 @@ public class RandomAccessFile implements DataOutput, DataInput
    * throws an exception if there is not enough data left in the stream to
    * fill the buffer
    *
-   * @param buf The buffer into which to read the data
+   * @param buffer The buffer into which to read the data
    *
    * @exception EOFException If end of file is reached before filling the 
    * buffer
@@ -417,9 +438,9 @@ public class RandomAccessFile implements DataOutput, DataInput
    * available and throws an exception if there is not enough data left in 
    * the stream to read <code>len</code> bytes.
    *
-   * @param buf The buffer into which to read the data
+   * @param buffer The buffer into which to read the data
    * @param offset The offset into the buffer to start storing data
-   * @param len The number of bytes to read into the buffer
+   * @param count The number of bytes to read into the buffer
    *
    * @exception EOFException If end of file is reached before filling 
    * the buffer
@@ -443,8 +464,8 @@ public class RandomAccessFile implements DataOutput, DataInput
    * four bytes read from the stream, they will be
    * transformed to an <code>int</code> in the following manner:
    * <p>
-   * <code>(int)(((byte1 & 0xFF) << 24) + ((byte2 & 0xFF) << 16) + 
-   * ((byte3 & 0xFF) << 8) + (byte4 & 0xFF)))</code>
+   * <code>(int)(((byte1 &amp; 0xFF) &lt;&lt; 24) + ((byte2 &amp; 0xFF) &lt;&lt; 16) + 
+   * ((byte3 &amp; 0xFF) &lt;&lt; 8) + (byte4 &amp; 0xFF)))</code>
    * <p>
    * The value returned is in the range of 0 to 65535.
    * <p>
@@ -507,10 +528,10 @@ public class RandomAccessFile implements DataOutput, DataInput
    * transformed to an <code>long</code> in the following manner:
    * <p>
    * <code>
-   * (long)((((long)byte1 & 0xFF) << 56) + (((long)byte2 & 0xFF) << 48) + 
-   * (((long)byte3 & 0xFF) << 40) + (((long)byte4 & 0xFF) << 32) + 
-   * (((long)byte5 & 0xFF) << 24) + (((long)byte6 & 0xFF) << 16) + 
-   * (((long)byte7 & 0xFF) << 8) + ((long)byte9 & 0xFF)))</code>
+   * (long)((((long)byte1 &amp; 0xFF) &lt;&lt; 56) + (((long)byte2 &amp; 0xFF) &lt;&lt; 48) + 
+   * (((long)byte3 &amp; 0xFF) &lt;&lt; 40) + (((long)byte4 &amp; 0xFF) &lt;&lt; 32) + 
+   * (((long)byte5 &amp; 0xFF) &lt;&lt; 24) + (((long)byte6 &amp; 0xFF) &lt;&lt; 16) + 
+   * (((long)byte7 &amp; 0xFF) &lt;&lt; 8) + ((long)byte9 &amp; 0xFF)))</code>
    * <p>
    * The value returned is in the range of 0 to 65535.
    * <p>
@@ -537,12 +558,12 @@ public class RandomAccessFile implements DataOutput, DataInput
    * significant byte first (i.e., "big endian") regardless of the native
    * host byte ordering. 
    * <p>
-   * As an example, if <code>byte1</code> and code{byte2</code> 
+   * As an example, if <code>byte1</code> and <code>byte2</code> 
    * represent the first
    * and second byte read from the stream respectively, they will be
    * transformed to a <code>short</code> in the following manner:
    * <p>
-   * <code>(short)(((byte1 & 0xFF) << 8) | (byte2 & 0xFF)</code>
+   * <code>(short)(((byte1 &amp; 0xFF) &lt;&lt; 8) | (byte2 &amp; 0xFF)</code>
    * <p>
    * The value returned is in the range of -32768 to 32767.
    * <p>
@@ -595,7 +616,7 @@ public class RandomAccessFile implements DataOutput, DataInput
    * and second byte read from the stream respectively, they will be
    * transformed to an <code>int</code> in the following manner:
    * <p>
-   * <code>(int)(((byte1 & 0xFF) << 8) + (byte2 & 0xFF))</code>
+   * <code>(int)(((byte1 &amp; 0xFF) &lt;&lt; 8) + (byte2 &amp; 0xFF))</code>
    * <p>
    * The value returned is in the range of 0 to 65535.
    * <p>
@@ -702,7 +723,7 @@ public class RandomAccessFile implements DataOutput, DataInput
    */
   public void seek (long pos) throws IOException
   {
-    fd.seek (pos, FileDescriptor.SET, false);
+    ch.position(pos);
   }
 
   /**
@@ -726,17 +747,20 @@ public class RandomAccessFile implements DataOutput, DataInput
     if (numBytes == 0)
       return 0;
     
-    long curPos = fd.getFilePointer ();
-    long newPos = fd.seek (numBytes, FileDescriptor.CUR, true);
-    
-    return (int) (newPos - curPos);
+    long oldPos = ch.position();
+    long newPos = oldPos + numBytes;
+    long size = ch.size();
+    if (newPos > size)
+      newPos = size;
+    ch.position(newPos);
+    return (int) (ch.position() - oldPos);
   }
 
   /**
    * This method writes a single byte of data to the file. The file must
    * be open for read-write in order for this operation to succeed.
    *
-   * @param The byte of data to write, passed as an int.
+   * @param oneByte The byte of data to write, passed as an int.
    *
    * @exception IOException If an error occurs
    */
@@ -749,7 +773,7 @@ public class RandomAccessFile implements DataOutput, DataInput
    * This method writes all the bytes in the specified array to the file.
    * The file must be open read-write in order for this operation to succeed.
    *
-   * @param buf The array of bytes to write to the file
+   * @param buffer The array of bytes to write to the file
    */
   public void write (byte[] buffer) throws IOException
   {
@@ -760,7 +784,7 @@ public class RandomAccessFile implements DataOutput, DataInput
    * This method writes <code>len</code> bytes to the file from the specified
    * array starting at index <code>offset</code> into the array.
    *
-   * @param buf The array of bytes to write to the file
+   * @param buffer The array of bytes to write to the file
    * @param offset The index into the array to start writing file
    * @param len The number of bytes to write
    *
@@ -776,7 +800,7 @@ public class RandomAccessFile implements DataOutput, DataInput
    * stream. For a value of <code>true</code>, 1 is written to the stream.
    * For a value of <code>false</code>, 0 is written.
    *
-   * @param b The <code>boolean</code> value to write to the stream
+   * @param val The <code>boolean</code> value to write to the stream
    *
    * @exception IOException If an error occurs
    */
@@ -789,68 +813,68 @@ public class RandomAccessFile implements DataOutput, DataInput
    * This method writes a Java <code>byte</code> value to the underlying
    * output stream.
    *
-   * @param b The <code>byte</code> to write to the stream, passed 
+   * @param val The <code>byte</code> to write to the stream, passed 
    * as an <code>int</code>.
    *
    * @exception IOException If an error occurs
    */
-  public final void writeByte (int v) throws IOException
+  public final void writeByte (int val) throws IOException
   {
-    out.writeByte(v);
+    out.writeByte(val);
   }
 
   /**
    * This method writes a Java <code>short</code> to the stream, high byte
    * first.  This method requires two bytes to encode the value.
    *
-   * @param s The <code>short</code> value to write to the stream, 
+   * @param val The <code>short</code> value to write to the stream, 
    * passed as an <code>int</code>.
    *
    * @exception IOException If an error occurs
    */
-  public final void writeShort (int v) throws IOException
+  public final void writeShort (int val) throws IOException
   {
-    out.writeShort(v);
+    out.writeShort(val);
   }
 
   /**
    * This method writes a single <code>char</code> value to the stream,
    * high byte first.
    *
-   * @param v The <code>char</code> value to write, passed as 
+   * @param val The <code>char</code> value to write, passed as 
    * an <code>int</code>.
    *
    * @exception IOException If an error occurs
    */
-  public final void writeChar (int v) throws IOException
+  public final void writeChar (int val) throws IOException
   {
-    out.writeChar(v);
+    out.writeChar(val);
   }
 
   /**
    * This method writes a Java <code>int</code> to the stream, high bytes
    * first.  This method requires four bytes to encode the value.
    *
-   * @param v The <code>int</code> value to write to the stream.
+   * @param val The <code>int</code> value to write to the stream.
    *
    * @exception IOException If an error occurs
    */
-  public final void writeInt (int v) throws IOException
+  public final void writeInt (int val) throws IOException
   {
-    out.writeInt(v);
+    out.writeInt(val);
   }
 
   /**
    * This method writes a Java <code>long</code> to the stream, high bytes
    * first.  This method requires eight bytes to encode the value.
    *
-   * @param v The <code>long</code> value to write to the stream.
+   * @param val The <code>long</code> value to write to the stream.
    *
    * @exception IOException If an error occurs
    */
-  public final void writeLong (long v) throws IOException
+  public final void writeLong (long val) throws IOException
   {
-    out.writeLong(v);
+    out.writeLong(val);
   }
 
   /**
@@ -861,15 +885,15 @@ public class RandomAccessFile implements DataOutput, DataInput
    * then writing this <code>int</code> value to the stream exactly the same
    * as the <code>writeInt()</code> method does.
    *
-   * @param v The floating point number to write to the stream.
+   * @param val The floating point number to write to the stream.
    *
    * @exception IOException If an error occurs
    *
    * @see #writeInt(int)
    */
-  public final void writeFloat (float v) throws IOException
+  public final void writeFloat (float val) throws IOException
   {
-    out.writeFloat(v);
+    out.writeFloat(val);
   }
 
   /**
@@ -880,16 +904,16 @@ public class RandomAccessFile implements DataOutput, DataInput
    * then writing this <code>long</code> value to the stream exactly the same
    * as the <code>writeLong()</code> method does.
    *
-   * @param v The double precision floating point number to write to the 
+   * @param val The double precision floating point number to write to the 
    * stream.
    *
    * @exception IOException If an error occurs
    *
    * @see #writeLong(long)
    */
-  public final void writeDouble (double v) throws IOException
+  public final void writeDouble (double val) throws IOException
   {
-    out.writeDouble(v);
+    out.writeDouble(val);
   }
 
   /**
@@ -897,13 +921,13 @@ public class RandomAccessFile implements DataOutput, DataInput
    * stream.  One byte is written for each character in the <code>String</code>.
    * The high eight bits of each character are discarded.
    *
-   * @param s The <code>String</code> to write to the stream
+   * @param val The <code>String</code> to write to the stream
    *
    * @exception IOException If an error occurs
    */
-  public final void writeBytes (String s) throws IOException
+  public final void writeBytes (String val) throws IOException
   {
-    out.writeBytes(s);
+    out.writeBytes(val);
   }
   
   /**
@@ -911,13 +935,13 @@ public class RandomAccessFile implements DataOutput, DataInput
    * stream.  There will be two bytes for each character value.  The high
    * byte of the character will be written first.
    *
-   * @param s The <code>String</code> to write to the stream.
+   * @param val The <code>String</code> to write to the stream.
    *
    * @exception IOException If an error occurs
    */
-  public final void writeChars (String s) throws IOException
+  public final void writeChars (String val) throws IOException
   {
-    out.writeChars(s);
+    out.writeChars(val);
   }
   
   /**
@@ -945,13 +969,13 @@ public class RandomAccessFile implements DataOutput, DataInput
    * character value are stored in bits 0-5 of byte three, with the high bits
    * of that byte set to "10".
    *
-   * @param s The <code>String</code> to write to the output in UTF format
+   * @param val The <code>String</code> to write to the output in UTF format
    *
    * @exception IOException If an error occurs
    */
-  public final void writeUTF (String s) throws IOException
+  public final void writeUTF (String val) throws IOException
   {
-    out.writeUTF(s);
+    out.writeUTF(val);
   }
   
   /**
@@ -962,11 +986,6 @@ public class RandomAccessFile implements DataOutput, DataInput
    */
   public final synchronized FileChannel getChannel ()
   {
-    if (ch == null)
-      ch = new FileChannelImpl (fd, true, this);
-
     return ch;
   }
-
-} // class RandomAccessFile
-
+}
